/*
Joshua Stough
Washington and Lee University
Quicksort a random list of size given by the argument (default 1M)
Time both sequential quicksort and parallel (using Pthreads).
See https://computing.llnl.gov/tutorials/pthreads/#Designing

$ make
$ ./parallelQuicksort2 --libc-qsort --parallel-qsort --sequential-qsort --seed 12345 --size 1000000 --output measurements_args.txt

I (Arnaud Legrand) downloaded this code from
http://sc12.supercomputing.org/hpceducator/PythonForParallelism/codes/parallelQuicksort.c
and fixed it thanks to my two friends Valgrind and gdb. I also
reindented and modified the commnad line management to better suit my
needs.
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/resource.h>

#define DNUM 1000000
#define THREAD_LEVEL 10
static int thread_level;

//for sequential and parallel implementation
void swap(double lyst[], int i, int j);
int partition(double lyst[], int lo, int hi);
void quicksortHelper(double lyst[], int lo, int hi);
void quicksort(double lyst[], int size);
int isSorted(double lyst[], int size);

//for parallel implementation
void parallelQuicksort(double lyst[], int size, int tlevel);
void *parallelQuicksortHelper(void *threadarg)  __attribute__ ((noreturn));
struct thread_data {
  double *lyst;
  int low;
  int high;
  int level;
};
//thread_data should be thread-safe, since while lyst is 
//shared, [low, high] will not overlap among threads.

//for the builtin qsort, for fun:
int compare_doubles(const void *a, const void *b);

static struct option long_options[] = {
  {"seed",             required_argument, 0,  'r' },
  {"size",             required_argument, 0,  's' },
  {"libc-qsort",       no_argument,       0,  'l' },
  {"parallel-qsort" ,  no_argument,       0,  'p' },
  {"sequential-qsort", no_argument,       0,  'q' },
  {"output",           required_argument, 0,  'o' },
  {"num-measurements", required_argument, 0,  'n' },
  {"thread-level",     required_argument, 0,  't' },
  {"help",             no_argument,       0,  'h' },
  {0,                  0,                 0,  0 }
};

struct parameters {
  int seed;
  int size;
  int libc_qsort;
  int parallel_qsort;
  int sequential_qsort;
  int num_measurements;
  int thread_level;
};

typedef enum {
  TEST_BUILTIN = 0,
  TEST_SEQUENTIAL,
  TEST_PARALLEL
} e_test_t;

static void print_usage(int argc, char *argv[]) {
  int i;
  printf("Usage: %s --option=<value>, where option is in the following list\n", argv[0]);
  for(i=0;long_options[i].name;i++) {
    printf("\t%s",long_options[i].name);
    if(long_options[i].has_arg == required_argument) 
      printf(" (requires a valid argument)");
    printf("\n");
  }
}

/*
Main method:
-generate random list
-time sequential quicksort
-time parallel quicksort
-time standard qsort
*/
int main(int argc, char *argv[])
{
  int opt;
  
  int option_index = 0;
  FILE *output = stdout;

  struct parameters parameters;
  parameters.size = DNUM;
  parameters.seed = 0;
  parameters.libc_qsort = 0;
  parameters.parallel_qsort = 0;
  parameters.sequential_qsort = 0;
  parameters.num_measurements = 1;
  parameters.thread_level = THREAD_LEVEL;
  
  while((opt = getopt_long(argc, argv, "", long_options, &option_index))!=-1) {
    switch (opt) {
    case 'r' : parameters.seed = atoi(optarg);
      break;
    case 's' : parameters.size = atoi(optarg);
      break;
    case 'l' : parameters.libc_qsort = 1;
      break;
    case 'p' : parameters.parallel_qsort = 1;
      break;
    case 'q' : parameters.sequential_qsort = 1;
      break;
    case 'n' : parameters.num_measurements = atoi(optarg);
      break;
    case 't' : parameters.thread_level = atoi(optarg);
      break;
    case 'o' : 
      if(!(output = fopen(optarg,"a"))) {
	fprintf(stderr,"Unable to open \"%s\"\n",optarg);
      }
      break;
    case 'h' :
    default: print_usage(argc,argv); 
      exit(EXIT_FAILURE);
    }
  }

  if(!parameters.seed) {
    parameters.seed = time(NULL);
  } 
  srand(parameters.seed);

  int NUM = parameters.size;
  thread_level = parameters.thread_level;

  {
    fprintf(output,"parameters.size: %d\n", parameters.size);
    fprintf(output,"parameters.seed: %d\n", parameters.seed);
    fprintf(output,"parameters.num_measurements: %d\n", parameters.num_measurements);
    fprintf(output,"parameters.libc_qsort: %d\n", parameters.libc_qsort);
    fprintf(output,"parameters.parallel_qsort: %d\n", parameters.parallel_qsort);
    fprintf(output,"parameters.thread_level: %d\n", parameters.thread_level);
    fprintf(output,"parameters.sequential_qsort: %d\n", parameters.sequential_qsort);

  }
  //Want to compare sorting on the same list,
  //so backup.
  double *lystbck = (double *) malloc(NUM * sizeof(double));
  double *lyst = (double *) malloc(NUM * sizeof(double));
  int i;

  //Populate random original/backup list.
  for (i = 0; i < NUM; i++) {
    lystbck[i] = 1.0 * rand() / RAND_MAX;
  }

  void test_sort_function(e_test_t test, const char *method_name) {
    struct timespec start, end;
    struct rusage rstart, rend;
    double diff, ru_utime, ru_stime;

    memcpy(lyst, lystbck, NUM * sizeof(double));
    getrusage(RUSAGE_SELF, &rstart);
    clock_gettime(CLOCK_REALTIME,&start);
    switch (test) {
    case TEST_BUILTIN : qsort(lyst, NUM, sizeof(double), compare_doubles);
      break;
    case TEST_SEQUENTIAL : quicksort(lyst, NUM);
      break;
    case TEST_PARALLEL : parallelQuicksort(lyst, NUM, thread_level);
      break;
    }
    clock_gettime(CLOCK_REALTIME,&end);
    getrusage(RUSAGE_SELF, &rend);
      
    if (!isSorted(lyst, NUM)) {
      fprintf(stderr,"Oops, lyst did not get sorted by %s.\n",method_name);
    }
    //Compute time difference.
    diff = ((end.tv_sec * 1e9 + end.tv_nsec)
	    - (start.tv_sec * 1e9 + start.tv_nsec)) / 1e9;
    ru_utime = ((rend.ru_utime.tv_sec * 1e6 + rend.ru_utime.tv_usec)
		- (rstart.ru_utime.tv_sec * 1e6 + rstart.ru_utime.tv_usec)) / 1e6;
    ru_stime = ((rend.ru_stime.tv_sec * 1e6 + rend.ru_stime.tv_usec)
		- (rstart.ru_stime.tv_sec * 1e6 + rstart.ru_stime.tv_usec)) / 1e6;
    fprintf(output,"%s took: %lf sec.\n", method_name, diff);
    fprintf(output,"%s user time: %lf sec.\n", method_name, ru_utime);
    fprintf(output,"%s sytem time: %lf sec.\n", method_name, ru_stime);
    fprintf(output,"%s PCPU: %lf %%.\n", method_name, 
	    100.0*(ru_stime+ru_utime)/diff);
  }

  if(parameters.sequential_qsort) { //Sequential mergesort, and timing.
    for(i = 0 ; i < parameters.num_measurements; i++) {
      test_sort_function(TEST_SEQUENTIAL,"Sequential quicksort");
    }
  }

  if(parameters.parallel_qsort) { // Now, parallel quicksort.
    for(i = 0 ; i < parameters.num_measurements; i++) {
      test_sort_function(TEST_PARALLEL,"Parallel quicksort");
    }
  }

  if(parameters.libc_qsort) {   //Finally, built-in for reference:
    for(i = 0 ; i < parameters.num_measurements; i++) {
      test_sort_function(TEST_BUILTIN,"Built-in quicksort");
    }
  }

  fclose(output);
  free(lyst);
  free(lystbck);
  pthread_exit(NULL);
}

void quicksort(double lyst[], int size)
{
  quicksortHelper(lyst, 0, size - 1);
}

void quicksortHelper(double lyst[], int lo, int hi)
{
  if (lo >= hi)
    return;
  int b = partition(lyst, lo, hi);
  quicksortHelper(lyst, lo, b - 1);
  quicksortHelper(lyst, b + 1, hi);
}

void swap(double lyst[], int i, int j)
{
  double temp = lyst[i];
  lyst[i] = lyst[j];
  lyst[j] = temp;
}

int partition(double lyst[], int lo, int hi)
{
  int b = lo;
  int r = (int) (lo + (hi - lo) * (1.0 * rand() / RAND_MAX));
  double pivot = lyst[r];
  int i;
  swap(lyst, r, hi);
  for (i = lo; i < hi; i++) {
    if (lyst[i] < pivot) {
      swap(lyst, i, b);
      b++;
    }
  }
  swap(lyst, hi, b);
  return b;
}


/*
parallel quicksort top level:
instantiate parallelQuicksortHelper thread, and that's 
basically it.
*/
void parallelQuicksort(double lyst[], int size, int tlevel)
{
  int rc;
  void *status;

  //Want joinable threads (usually default).
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  //pthread function can take only one argument, so struct.
  struct thread_data td;
  td.lyst = lyst;
  td.low = 0;
  td.high = size - 1;
  td.level = tlevel;

  //The top-level thread.
  pthread_t theThread;
  rc = pthread_create(&theThread, &attr, parallelQuicksortHelper,
		      (void *) &td);
  if (rc) {
    fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
    exit(-1);
  }
  //Now join the thread (wait, as joining blocks) and quit.
  pthread_attr_destroy(&attr);
  rc = pthread_join(theThread, &status);
  if (rc) {
    fprintf(stderr,"ERROR; return code from pthread_join() is %d\n", rc);
    exit(-1);
  }
  //printf("Main: completed join with top thread having a status of %ld\n",
  //              (long)status);

}

/*
parallelQuicksortHelper
-if the level is still > 0, then partition and make 
parallelQuicksortHelper threads to solve the left and 
right-hand sides, then quit. Otherwise, call sequential.
*/
void *parallelQuicksortHelper(void *threadarg)
{
  int mid, t, rc;
  void *status;

  struct thread_data *my_data;
  my_data = (struct thread_data *) threadarg;

  //fyi:
  //printf("Thread responsible for [%d, %d], level %d.\n",
  //              my_data->low, my_data->high, my_data->level);

  if (my_data->level <= 0 || my_data->low == my_data->high + 1) {
    //We have plenty of threads, finish with sequential.
    quicksortHelper(my_data->lyst, my_data->low, my_data->high);
    pthread_exit(NULL);
  }
  //Want joinable threads (usually default).
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  //Now we partition our part of the lyst.
  mid = partition(my_data->lyst, my_data->low, my_data->high);

  //At this point, we will create threads for the 
  //left and right sides.  Must create their data args.
  struct thread_data thread_data_array[2];

  for (t = 0; t < 2; t++) {
    thread_data_array[t].lyst = my_data->lyst;
    thread_data_array[t].level = my_data->level - 1;
  }
  thread_data_array[0].low = my_data->low;
  thread_data_array[0].high = mid - 1;
  thread_data_array[1].low = mid + 1;
  thread_data_array[1].high = my_data->high;

  //Now, instantiate the threads.
  //In quicksort of course, due to the transitive property,
  //no elements in the left and right sides of mid will have
  //to be compared again.
  pthread_t threads[2];
  for (t = 0; t < 2; t++) {
    rc = pthread_create(&threads[t], &attr, parallelQuicksortHelper,
			(void *) &thread_data_array[t]);
    if (rc) {
      fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
      exit(-1);
    }
  }

  pthread_attr_destroy(&attr);
  //Now, join the left and right sides to finish.
  for (t = 0; t < 2; t++) {
    rc = pthread_join(threads[t], &status);
    if (rc) {
      fprintf(stderr,"ERROR; return code from pthread_join() is %d\n", rc);
      exit(-1);
    }
  }

  pthread_exit(NULL);
}

//check if the elements of lyst are in non-decreasing order.
//one is success.
int isSorted(double lyst[], int size)
{
  int i;
  for (i = 1; i < size; i++) {
    if (lyst[i] < lyst[i - 1]) {
      fprintf(stderr,"at loc %d, %e < %e \n", i, lyst[i], lyst[i - 1]);
      return 0;
    }
  }
  return 1;
}

//for the built-in qsort comparator
//from http://www.gnu.org/software/libc/manual/html_node/Comparison-Functions.html#Comparison-Functions
int compare_doubles(const void *a, const void *b)
{
  const double *da = (const double *) a;
  const double *db = (const double *) b;

  return (*da > *db) - (*da < *db);
}
